@@ -8,7 +8,7 @@ QT_BEGIN_NAMESPACE
 
 /*!
     \qmltype DelegateChoice
-//!    \instantiates QQmlDelegateChoice
+//!    \nativetype QQmlDelegateChoice
     \inqmlmodule Qt.labs.qmlmodels
     \brief Encapsulates a delegate and when to use it.
 
@@ -21,7 +21,7 @@ QT_BEGIN_NAMESPACE
 */
 
 /*!
-    \qmlproperty variant QtQml.Models::DelegateChoice::roleValue
+    \qmlproperty variant Qt.labs.qmlmodels::DelegateChoice::roleValue
     This property holds the value used to match the role data for the role provided by \l DelegateChooser::role.
 */
 QVariant QQmlDelegateChoice::roleValue() const
@@ -39,7 +39,7 @@ void QQmlDelegateChoice::setRoleValue(const QVariant &value)
 }
 
 /*!
-    \qmlproperty int QtQml.Models::DelegateChoice::row
+    \qmlproperty int Qt.labs.qmlmodels::DelegateChoice::row
     This property holds the value used to match the row value of model elements.
     With models that have only the index property (and thus only one column), this property
     should be intended as an index, and set to the desired index value.
@@ -51,7 +51,7 @@ void QQmlDelegateChoice::setRoleValue(const QVariant &value)
 */
 
 /*!
-    \qmlproperty int QtQml.Models::DelegateChoice::index
+    \qmlproperty int Qt.labs.qmlmodels::DelegateChoice::index
     This property holds the value used to match the index value of model elements.
     This is effectively an alias for \l row.
 
@@ -73,7 +73,7 @@ void QQmlDelegateChoice::setRow(int r)
 }
 
 /*!
-    \qmlproperty int QtQml.Models::DelegateChoice::column
+    \qmlproperty int Qt.labs.qmlmodels::DelegateChoice::column
     This property holds the value used to match the column value of model elements.
 */
 int QQmlDelegateChoice::column() const
@@ -96,7 +96,7 @@ QQmlComponent *QQmlDelegateChoice::delegate() const
 }
 
 /*!
-    \qmlproperty Component QtQml.Models::DelegateChoice::delegate
+    \qmlproperty Component Qt.labs.qmlmodels::DelegateChoice::delegate
     This property holds the delegate to use if this choice matches the model item.
 */
 void QQmlDelegateChoice::setDelegate(QQmlComponent *delegate)
@@ -137,7 +137,7 @@ bool QQmlDelegateChoice::match(int row, int column, const QVariant &value) const
 
 /*!
     \qmltype DelegateChooser
-//!    \instantiates QQmlDelegateChooser
+//!    \nativetype QQmlDelegateChooser
     \inqmlmodule Qt.labs.qmlmodels
     \brief Allows a view to use different delegates for different types of items in the model.
 
@@ -190,9 +190,13 @@ bool QQmlDelegateChoice::match(int row, int column, const QVariant &value) const
 */
 
 /*!
-    \qmlproperty string QtQml.Models::DelegateChooser::role
+    \qmlproperty string Qt.labs.qmlmodels::DelegateChooser::role
     This property holds the role or the property name used to determine the delegate for a given model item.
 
+    \note For \l{QAbstractItemModel} based models, including \l{ListModel}, the DelegateChooser will
+    reevaluate the choice when the model signals that the role has changed. For any other type of model,
+    this choice will only be done once when the item for a given model index is created.
+
     \sa DelegateChoice
 */
 void QQmlDelegateChooser::setRole(const QString &role)
@@ -204,7 +208,7 @@ void QQmlDelegateChooser::setRole(const QString &role)
 }
 
 /*!
-    \qmlproperty list<DelegateChoice> QtQml.Models::DelegateChooser::choices
+    \qmlproperty list<DelegateChoice> Qt.labs.qmlmodels::DelegateChooser::choices
     \qmldefault
 
     The list of DelegateChoices for the chooser.
@@ -233,7 +237,7 @@ void QQmlDelegateChooser::choices_append(QQmlListProperty<QQmlDelegateChoice> *p
     QQmlDelegateChooser *q = static_cast<QQmlDelegateChooser *>(prop->object);
     q->m_choices.append(choice);
     connect(choice, &QQmlDelegateChoice::changed, q, &QQmlAbstractDelegateComponent::delegateChanged);
-    q->delegateChanged();
+    emit q->delegateChanged();
 }
 
 qsizetype QQmlDelegateChooser::choices_count(QQmlListProperty<QQmlDelegateChoice> *prop)
@@ -254,7 +258,7 @@ void QQmlDelegateChooser::choices_clear(QQmlListProperty<QQmlDelegateChoice> *pr
     for (QQmlDelegateChoice *choice : q->m_choices)
         disconnect(choice, &QQmlDelegateChoice::changed, q, &QQmlAbstractDelegateComponent::delegateChanged);
     q->m_choices.clear();
-    q->delegateChanged();
+    emit q->delegateChanged();
 }
 
 void QQmlDelegateChooser::choices_replace(QQmlListProperty<QQmlDelegateChoice> *prop,
@@ -266,7 +270,7 @@ void QQmlDelegateChooser::choices_replace(QQmlListProperty<QQmlDelegateChoice> *
     q->m_choices[index] = choice;
     connect(choice, &QQmlDelegateChoice::changed, q,
             &QQmlAbstractDelegateComponent::delegateChanged);
-    q->delegateChanged();
+    emit q->delegateChanged();
 }
 
 void QQmlDelegateChooser::choices_removeLast(QQmlListProperty<QQmlDelegateChoice> *prop)
@@ -274,7 +278,7 @@ void QQmlDelegateChooser::choices_removeLast(QQmlListProperty<QQmlDelegateChoice
     QQmlDelegateChooser *q = static_cast<QQmlDelegateChooser *>(prop->object);
     disconnect(q->m_choices.takeLast(), &QQmlDelegateChoice::changed,
                q, &QQmlAbstractDelegateComponent::delegateChanged);
-    q->delegateChanged();
+    emit q->delegateChanged();
 }
 
 QQmlComponent *QQmlDelegateChooser::delegate(QQmlAdaptorModel *adaptorModel, int row, int column) const
