@@ -46,6 +46,9 @@ namespace {
     };
 }
 
+Q_TRACE_POINT(qtqml, QQmlCompiling_entry, const QUrl &url)
+Q_TRACE_POINT(qtqml, QQmlCompiling_exit)
+
 /*!
 \class QQmlTypeLoader
 \brief The QQmlTypeLoader class abstracts loading files and their dependencies over the network.
@@ -88,8 +91,6 @@ void QQmlTypeLoader::invalidate()
     // Need to delete the network replies after
     // the loader thread is shutdown as it could be
     // getting new replies while we clear them
-    for (NetworkReplies::Iterator iter = m_networkReplies.begin(); iter != m_networkReplies.end(); ++iter)
-        (*iter)->release();
     m_networkReplies.clear();
 #endif // qml_network
 }
@@ -181,9 +182,7 @@ void QQmlTypeLoader::doLoad(const Loader &loader, QQmlDataBlob *blob, Mode mode)
         } else {
             Q_ASSERT(mode == Synchronous);
             while (!blob->isCompleteOrError()) {
-                unlock();
                 m_thread->waitForNextMessage();
-                lock();
             }
         }
     }
@@ -214,21 +213,28 @@ void QQmlTypeLoader::loadWithCachedUnit(QQmlDataBlob *blob, const QQmlPrivate::C
     doLoad(CachedLoader(unit), blob, mode);
 }
 
-void QQmlTypeLoader::loadWithStaticDataThread(QQmlDataBlob *blob, const QByteArray &data)
+void QQmlTypeLoader::drop(const QQmlDataBlob::Ptr &blob)
+{
+    // We must not destroy a QQmlDataBlob from the main thread
+    // since it will shuffle its dependencies around.
+    m_thread->drop(blob);
+}
+
+void QQmlTypeLoader::loadWithStaticDataThread(const QQmlDataBlob::Ptr &blob, const QByteArray &data)
 {
     ASSERT_LOADTHREAD();
 
     setData(blob, data);
 }
 
-void QQmlTypeLoader::loadWithCachedUnitThread(QQmlDataBlob *blob, const QQmlPrivate::CachedQmlUnit *unit)
+void QQmlTypeLoader::loadWithCachedUnitThread(const QQmlDataBlob::Ptr &blob, const QQmlPrivate::CachedQmlUnit *unit)
 {
     ASSERT_LOADTHREAD();
 
     setCachedUnit(blob, unit);
 }
 
-void QQmlTypeLoader::loadThread(QQmlDataBlob *blob)
+void QQmlTypeLoader::loadThread(const QQmlDataBlob::Ptr &blob)
 {
     ASSERT_LOADTHREAD();
 
@@ -254,7 +260,7 @@ void QQmlTypeLoader::loadThread(QQmlDataBlob *blob)
             return;
         }
 
-        blob->m_data.setProgress(0xFF);
+        blob->m_data.setProgress(1.f);
         if (blob->m_data.isAsync())
             m_thread->callDownloadProgressChanged(blob, 1.);
 
@@ -264,7 +270,6 @@ void QQmlTypeLoader::loadThread(QQmlDataBlob *blob)
 #if QT_CONFIG(qml_network)
         QNetworkReply *reply = m_thread->networkAccessManager()->get(QNetworkRequest(blob->m_url));
         QQmlTypeLoaderNetworkReplyProxy *nrp = m_thread->networkReplyProxy();
-        blob->addref();
         m_networkReplies.insert(reply, blob);
 
         if (reply->isFinished()) {
@@ -296,7 +301,7 @@ void QQmlTypeLoader::networkReplyFinished(QNetworkReply *reply)
 
     reply->deleteLater();
 
-    QQmlDataBlob *blob = m_networkReplies.take(reply);
+    QQmlRefPointer<QQmlDataBlob> blob = m_networkReplies.take(reply);
 
     Q_ASSERT(blob);
 
@@ -312,7 +317,7 @@ void QQmlTypeLoader::networkReplyFinished(QNetworkReply *reply)
             QNetworkReply *reply = m_thread->networkAccessManager()->get(QNetworkRequest(url));
             QObject *nrp = m_thread->networkReplyProxy();
             QObject::connect(reply, SIGNAL(finished()), nrp, SLOT(finished()));
-            m_networkReplies.insert(reply, blob);
+            m_networkReplies.insert(reply, std::move(blob));
 #ifdef DATABLOB_DEBUG
             qWarning("QQmlDataBlob: redirected to %s", qPrintable(blob->finalUrlString()));
 #endif
@@ -326,8 +331,6 @@ void QQmlTypeLoader::networkReplyFinished(QNetworkReply *reply)
         QByteArray data = reply->readAll();
         setData(blob, data);
     }
-
-    blob->release();
 }
 
 void QQmlTypeLoader::networkReplyProgress(QNetworkReply *reply,
@@ -335,12 +338,12 @@ void QQmlTypeLoader::networkReplyProgress(QNetworkReply *reply,
 {
     Q_ASSERT(m_thread->isThisThread());
 
-    QQmlDataBlob *blob = m_networkReplies.value(reply);
+    const QQmlRefPointer<QQmlDataBlob> blob = m_networkReplies.value(reply);
 
     Q_ASSERT(blob);
 
     if (bytesTotal != 0) {
-        quint8 progress = 0xFF * (qreal(bytesReceived) / qreal(bytesTotal));
+        qreal progress = (qreal(bytesReceived) / qreal(bytesTotal));
         blob->m_data.setProgress(progress);
         if (blob->m_data.isAsync())
             m_thread->callDownloadProgressChanged(blob, blob->m_data.progress());
@@ -384,7 +387,7 @@ void QQmlTypeLoader::initializeEngine(QQmlExtensionInterface *iface, const char
     doInitializeEngine(iface, m_thread, engine(), uri);
 }
 
-void QQmlTypeLoader::setData(QQmlDataBlob *blob, const QByteArray &data)
+void QQmlTypeLoader::setData(const QQmlDataBlob::Ptr &blob, const QByteArray &data)
 {
     QQmlDataBlob::SourceCodeData d;
     d.inlineSourceCode = QString::fromUtf8(data);
@@ -392,17 +395,17 @@ void QQmlTypeLoader::setData(QQmlDataBlob *blob, const QByteArray &data)
     setData(blob, d);
 }
 
-void QQmlTypeLoader::setData(QQmlDataBlob *blob, const QString &fileName)
+void QQmlTypeLoader::setData(const QQmlDataBlob::Ptr &blob, const QString &fileName)
 {
     QQmlDataBlob::SourceCodeData d;
     d.fileInfo = QFileInfo(fileName);
     setData(blob, d);
 }
 
-void QQmlTypeLoader::setData(QQmlDataBlob *blob, const QQmlDataBlob::SourceCodeData &d)
+void QQmlTypeLoader::setData(const QQmlDataBlob::Ptr &blob, const QQmlDataBlob::SourceCodeData &d)
 {
     Q_TRACE_SCOPE(QQmlCompiling, blob->url());
-    QQmlCompilingProfiler prof(profiler(), blob);
+    QQmlCompilingProfiler prof(profiler(), blob.data());
 
     blob->m_inCallback = true;
 
@@ -419,10 +422,10 @@ void QQmlTypeLoader::setData(QQmlDataBlob *blob, const QQmlDataBlob::SourceCodeD
     blob->tryDone();
 }
 
-void QQmlTypeLoader::setCachedUnit(QQmlDataBlob *blob, const QQmlPrivate::CachedQmlUnit *unit)
+void QQmlTypeLoader::setCachedUnit(const QQmlDataBlob::Ptr &blob, const QQmlPrivate::CachedQmlUnit *unit)
 {
     Q_TRACE_SCOPE(QQmlCompiling, blob->url());
-    QQmlCompilingProfiler prof(profiler(), blob);
+    QQmlCompilingProfiler prof(profiler(), blob.data());
 
     blob->m_inCallback = true;
 
@@ -459,7 +462,7 @@ QQmlTypeLoader::Blob::PendingImport::PendingImport(
 
 QQmlTypeLoader::Blob::Blob(const QUrl &url, QQmlDataBlob::Type type, QQmlTypeLoader *loader)
   : QQmlDataBlob(url, type, loader)
-  , m_importCache(new QQmlImports(loader), QQmlRefPointer<QQmlImports>::Adopt)
+  , m_importCache(new QQmlImports(), QQmlRefPointer<QQmlImports>::Adopt)
 {
 }
 
@@ -486,7 +489,47 @@ bool QQmlTypeLoader::Blob::fetchQmldir(const QUrl &url, PendingImportPtr import,
     return true;
 }
 
-bool QQmlTypeLoader::Blob::updateQmldir(const QQmlRefPointer<QQmlQmldirData> &data, PendingImportPtr import, QList<QQmlError> *errors)
+/*!
+ * \internal
+ * Import any qualified scripts of for \a import as listed in \a qmldir.
+ * Precondition is that \a import is actually qualified.
+ */
+void QQmlTypeLoader::Blob::importQmldirScripts(
+        const QQmlTypeLoader::Blob::PendingImportPtr &import,
+        const QQmlTypeLoaderQmldirContent &qmldir, const QUrl &qmldirUrl)
+{
+    const auto qmldirScripts = qmldir.scripts();
+    for (const QQmlDirParser::Script &script : qmldirScripts) {
+        const QUrl scriptUrl = qmldirUrl.resolved(QUrl(script.fileName));
+        QQmlRefPointer<QQmlScriptBlob> blob = typeLoader()->getScript(scriptUrl);
+        addDependency(blob.data());
+        scriptImported(blob, import->location, script.nameSpace, import->qualifier);
+    }
+}
+
+template<typename URL>
+void postProcessQmldir(
+        QQmlTypeLoader::Blob *self,
+        const QQmlTypeLoader::Blob::PendingImportPtr &import, const QString &qmldirFilePath,
+        const URL &qmldirUrl)
+{
+    const QQmlTypeLoaderQmldirContent qmldir = self->typeLoader()->qmldirContent(qmldirFilePath);
+    if (!import->qualifier.isEmpty())
+        self->importQmldirScripts(import, qmldir, QUrl(qmldirUrl));
+
+    if (qmldir.plugins().isEmpty()) {
+        // If the qmldir does not register a plugin, we might still have declaratively
+        // registered types (if we are dealing with an application instead of a library)
+        // We should use module name given in the qmldir rather than the one given by the
+        // import since the import may be a directory import.
+        auto module = QQmlMetaType::typeModule(qmldir.typeNamespace(), import->version);
+        if (!module)
+            QQmlMetaType::qmlRegisterModuleTypes(qmldir.typeNamespace());
+        // else: If the module already exists, the types must have been already registered
+    }
+}
+
+bool QQmlTypeLoader::Blob::updateQmldir(const QQmlRefPointer<QQmlQmldirData> &data, const QQmlTypeLoader::Blob::PendingImportPtr &import, QList<QQmlError> *errors)
 {
     QString qmldirIdentifier = data->urlString();
     QString qmldirUrl = qmldirIdentifier.left(qmldirIdentifier.lastIndexOf(QLatin1Char('/')) + 1);
@@ -494,8 +537,7 @@ bool QQmlTypeLoader::Blob::updateQmldir(const QQmlRefPointer<QQmlQmldirData> &da
     typeLoader()->setQmldirContent(qmldirIdentifier, data->content());
 
     const QTypeRevision version = m_importCache->updateQmldirContent(
-                typeLoader()->importDatabase(), import->uri, import->qualifier, qmldirIdentifier,
-                qmldirUrl, errors);
+            typeLoader(), import->uri, import->qualifier, qmldirIdentifier, qmldirUrl, errors);
     if (!version.isValid())
         return false;
 
@@ -511,35 +553,26 @@ bool QQmlTypeLoader::Blob::updateQmldir(const QQmlRefPointer<QQmlQmldirData> &da
     // Release this reference at destruction
     m_qmldirs << data;
 
-    if (!import->qualifier.isEmpty()) {
-        // Does this library contain any qualified scripts?
-        QUrl libraryUrl(qmldirUrl);
-        const QQmlTypeLoaderQmldirContent qmldir = typeLoader()->qmldirContent(qmldirIdentifier);
-        const auto qmldirScripts = qmldir.scripts();
-        for (const QQmlDirParser::Script &script : qmldirScripts) {
-            QUrl scriptUrl = libraryUrl.resolved(QUrl(script.fileName));
-            QQmlRefPointer<QQmlScriptBlob> blob = typeLoader()->getScript(scriptUrl);
-            addDependency(blob.data());
-
-            scriptImported(blob, import->location, script.nameSpace, import->qualifier);
-        }
-    }
-
+    postProcessQmldir(this, import, qmldirIdentifier, qmldirUrl);
     return true;
 }
 
-bool QQmlTypeLoader::Blob::addScriptImport(PendingImportPtr import)
+bool QQmlTypeLoader::Blob::addScriptImport(const QQmlTypeLoader::Blob::PendingImportPtr &import)
 {
-    QUrl scriptUrl = finalUrl().resolved(QUrl(import->uri));
-    QQmlRefPointer<QQmlScriptBlob> blob = typeLoader()->getScript(scriptUrl);
+    const QUrl url(import->uri);
+    QQmlTypeLoader *loader = typeLoader();
+    QQmlRefPointer<QQmlScriptBlob> blob = loader->injectedScript(url);
+    if (!blob)
+        blob = loader->getScript(finalUrl().resolved(url));
+    else
+        Q_ASSERT(blob->status() == QQmlDataBlob::Status::Complete);
     addDependency(blob.data());
     scriptImported(blob, import->location, import->qualifier, QString());
     return true;
 }
 
-bool QQmlTypeLoader::Blob::addFileImport(PendingImportPtr import, QList<QQmlError> *errors)
+bool QQmlTypeLoader::Blob::addFileImport(const QQmlTypeLoader::Blob::PendingImportPtr &import, QList<QQmlError> *errors)
 {
-    QQmlImportDatabase *importDatabase = typeLoader()->importDatabase();
     QQmlImports::ImportFlags flags;
 
     QUrl importUrl(import->uri);
@@ -553,8 +586,8 @@ bool QQmlTypeLoader::Blob::addFileImport(PendingImportPtr import, QList<QQmlErro
     }
 
     const QTypeRevision version = m_importCache->addFileImport(
-                importDatabase, import->uri, import->qualifier, import->version, flags,
-                nullptr, errors);
+            typeLoader(), import->uri, import->qualifier, import->version, flags,
+            import->precedence, nullptr, errors);
     if (!version.isValid())
         return false;
 
@@ -562,14 +595,43 @@ bool QQmlTypeLoader::Blob::addFileImport(PendingImportPtr import, QList<QQmlErro
     if (version.hasMajorVersion())
         import->version = version;
 
-    if (flags & QQmlImports::ImportIncomplete)
+    if (flags & QQmlImports::ImportIncomplete) {
         if (!fetchQmldir(qmldirUrl, import, 1, errors))
             return false;
+    } else {
+        const QString qmldirFilePath = QQmlFile::urlToLocalFileOrQrc(qmldirUrl);
+        if (!loadImportDependencies(import, qmldirFilePath, import->flags, errors))
+            return false;
+
+        postProcessQmldir(this, import, qmldirFilePath, qmldirUrl);
+    }
 
     return true;
 }
 
-bool QQmlTypeLoader::Blob::addLibraryImport(PendingImportPtr import, QList<QQmlError> *errors)
+static void addDependencyImportError(
+        const QQmlTypeLoader::Blob::PendingImportPtr &import, QList<QQmlError> *errors)
+{
+    QQmlError error;
+    QString reason = errors->front().description();
+    if (reason.size() > 512)
+        reason = reason.first(252) + QLatin1String("... ...") + reason.last(252);
+    if (import->version.hasMajorVersion()) {
+        error.setDescription(QQmlImportDatabase::tr(
+                                     "module \"%1\" version %2.%3 cannot be imported because:\n%4")
+                                     .arg(import->uri).arg(import->version.majorVersion())
+                                     .arg(import->version.hasMinorVersion()
+                                                  ? QString::number(import->version.minorVersion())
+                                                  : QLatin1String("x"))
+                                     .arg(reason));
+    } else {
+        error.setDescription(QQmlImportDatabase::tr("module \"%1\" cannot be imported because:\n%2")
+                                     .arg(import->uri, reason));
+    }
+    errors->prepend(error);
+}
+
+bool QQmlTypeLoader::Blob::addLibraryImport(const QQmlTypeLoader::Blob::PendingImportPtr &import, QList<QQmlError> *errors)
 {
     QQmlImportDatabase *importDatabase = typeLoader()->importDatabase();
 
@@ -584,8 +646,8 @@ bool QQmlTypeLoader::Blob::addLibraryImport(PendingImportPtr import, QList<QQmlE
                 [&](const QString &qmldirFilePath, const QString &qmldirUrl) {
         // This is a local library import
         const QTypeRevision actualVersion = m_importCache->addLibraryImport(
-                    importDatabase, import->uri, import->qualifier,
-                    import->version, qmldirFilePath, qmldirUrl, import->flags, errors);
+                typeLoader(), import->uri, import->qualifier, import->version, qmldirFilePath,
+                qmldirUrl, import->flags, import->precedence, errors);
         if (!actualVersion.isValid())
             return false;
 
@@ -593,64 +655,48 @@ bool QQmlTypeLoader::Blob::addLibraryImport(PendingImportPtr import, QList<QQmlE
         if (actualVersion.hasMajorVersion())
             import->version = actualVersion;
 
-        if (!loadImportDependencies(
-                    import, qmldirFilePath,
-                    (import->flags & QQmlImports::ImportImplicit)
-                        ? QQmlImports::ImportImplicit
-                        : QQmlImports::ImportLowPrecedence,
-                    errors))
+        if (!loadImportDependencies(import, qmldirFilePath, import->flags, errors)) {
+            addDependencyImportError(import, errors);
             return false;
-
-        const QQmlTypeLoaderQmldirContent qmldir = typeLoader()->qmldirContent(qmldirFilePath);
-        if (!import->qualifier.isEmpty()) {
-            // Does this library contain any qualified scripts?
-            QUrl libraryUrl(qmldirUrl);
-            const auto qmldirScripts = qmldir.scripts();
-            for (const QQmlDirParser::Script &script : qmldirScripts) {
-                QUrl scriptUrl = libraryUrl.resolved(QUrl(script.fileName));
-                QQmlRefPointer<QQmlScriptBlob> blob = typeLoader()->getScript(scriptUrl);
-                addDependency(blob.data());
-
-                scriptImported(blob, import->location, script.nameSpace, import->qualifier);
-            }
-        }
-        if (!qmldir.plugins().size()) {
-            // If the qmldir does not register a plugin, we might still have declaratively
-            // registered types (if we are dealing with an application instead of a library)
-            auto module = QQmlMetaType::typeModule(import->uri, import->version);
-            // If the module already exists, the types must have been already registered
-            if (!module)
-                QQmlMetaType::qmlRegisterModuleTypes(import->uri);
         }
+
+        postProcessQmldir(this, import, qmldirFilePath, qmldirUrl);
         return true;
     });
 
     switch (qmldirResult) {
     case QQmlImportDatabase::QmldirFound:
         return true;
-    case QQmlImportDatabase::QmldirNotFound:
+    case QQmlImportDatabase::QmldirNotFound: {
+        if (!loadImportDependencies(import, QString(), import->flags, errors)) {
+            addDependencyImportError(import, errors);
+            return false;
+        }
+        break;
+    }
     case QQmlImportDatabase::QmldirInterceptedToRemote:
         break;
     case QQmlImportDatabase::QmldirRejected:
         return false;
     }
 
-    if (
+    // If there is a qmldir we cannot see, yet, then we have to wait.
+    // The qmldir might contain import directives.
+    if (qmldirResult != QQmlImportDatabase::QmldirInterceptedToRemote && (
             // Major version of module already registered:
             // We believe that the registration is complete.
             QQmlMetaType::typeModule(import->uri, import->version)
 
             // Otherwise, try to register further module types.
-            || (qmldirResult != QQmlImportDatabase::QmldirInterceptedToRemote
-                && QQmlMetaType::qmlRegisterModuleTypes(import->uri))
+            || QQmlMetaType::qmlRegisterModuleTypes(import->uri)
 
             // Otherwise, there is no way to register any further types.
             // Try with any module of that name.
-            || QQmlMetaType::latestModuleVersion(import->uri).isValid()) {
+            || QQmlMetaType::latestModuleVersion(import->uri).isValid())) {
 
         if (!m_importCache->addLibraryImport(
-                    importDatabase, import->uri, import->qualifier, import->version,
-                    QString(), QString(), import->flags, errors).isValid()) {
+                    typeLoader(), import->uri, import->qualifier, import->version, QString(),
+                    QString(), import->flags, import->precedence, errors).isValid()) {
             return false;
         }
     } else {
@@ -669,9 +715,9 @@ bool QQmlTypeLoader::Blob::addLibraryImport(PendingImportPtr import, QList<QQmlE
         if (!remotePathList.isEmpty()) {
             // Add this library and request the possible locations for it
             const QTypeRevision version = m_importCache->addLibraryImport(
-                        importDatabase, import->uri, import->qualifier, import->version,
-                        QString(), QString(), import->flags | QQmlImports::ImportIncomplete,
-                        errors);
+                    typeLoader(), import->uri, import->qualifier, import->version, QString(),
+                    QString(), import->flags | QQmlImports::ImportIncomplete, import->precedence,
+                    errors);
 
             if (!version.isValid())
                 return false;
@@ -724,12 +770,10 @@ bool QQmlTypeLoader::Blob::addImport(
     case QV4::CompiledData::Import::ImportScript:
         return addScriptImport(import);
     case QV4::CompiledData::Import::ImportInlineComponent:
-        Q_UNREACHABLE(); // addImport is never called with an inline component import
-        return false;
+        Q_UNREACHABLE_RETURN(false); // addImport is never called with an inline component import
     }
 
-    Q_UNREACHABLE();
-    return false;
+    Q_UNREACHABLE_RETURN(false);
 }
 
 void QQmlTypeLoader::Blob::dependencyComplete(QQmlDataBlob *blob)
@@ -752,7 +796,8 @@ void QQmlTypeLoader::Blob::dependencyComplete(QQmlDataBlob *blob)
 
 bool QQmlTypeLoader::Blob::loadDependentImports(
         const QList<QQmlDirParser::Import> &imports, const QString &qualifier,
-        QTypeRevision version, QQmlImports::ImportFlags flags, QList<QQmlError> *errors)
+        QTypeRevision version, quint16 precedence, QQmlImports::ImportFlags flags,
+        QList<QQmlError> *errors)
 {
     for (const auto &import : imports) {
         if (import.flags & QQmlDirParser::Import::Optional)
@@ -763,6 +808,7 @@ bool QQmlTypeLoader::Blob::loadDependentImports(
         dependencyImport->version = (import.flags & QQmlDirParser::Import::Auto)
                 ? version : import.version;
         dependencyImport->flags = flags;
+        dependencyImport->precedence = precedence;
 
         qCDebug(lcQmlImport)
                 << "loading dependent import" << dependencyImport->uri << "version"
@@ -785,16 +831,34 @@ bool QQmlTypeLoader::Blob::loadDependentImports(
 }
 
 bool QQmlTypeLoader::Blob::loadImportDependencies(
-        PendingImportPtr currentImport, const QString &qmldirUri, QQmlImports::ImportFlags flags,
-        QList<QQmlError> *errors)
+        const QQmlTypeLoader::Blob::PendingImportPtr &currentImport, const QString &qmldirUri,
+        QQmlImports::ImportFlags flags, QList<QQmlError> *errors)
 {
-    const QQmlTypeLoaderQmldirContent qmldir = typeLoader()->qmldirContent(qmldirUri);
-    const QList<QQmlDirParser::Import> implicitImports
-            = QQmlMetaType::moduleImports(currentImport->uri, currentImport->version)
-            + qmldir.imports();
+    QList<QQmlDirParser::Import> implicitImports
+            = QQmlMetaType::moduleImports(currentImport->uri, currentImport->version);
+    if (!qmldirUri.isEmpty())
+        implicitImports += typeLoader()->qmldirContent(qmldirUri).imports();
+
+    // Prevent overflow from one category of import into the other.
+    switch (currentImport->precedence) {
+    case QQmlImportInstance::Implicit - 1:
+    case QQmlImportInstance::Lowest: {
+        QQmlError error;
+        error.setDescription(
+                    QString::fromLatin1("Too many dependent imports for %1 %2.%3")
+                    .arg(currentImport->uri)
+                    .arg(currentImport->version.majorVersion())
+                    .arg(currentImport->version.minorVersion()));
+        errors->append(error);
+        return false;
+    }
+    default:
+        break;
+    }
+
     if (!loadDependentImports(
                 implicitImports, currentImport->qualifier, currentImport->version,
-                flags, errors)) {
+                currentImport->precedence + 1, flags, errors)) {
         QQmlError error;
         error.setDescription(
                     QString::fromLatin1(
@@ -814,9 +878,27 @@ bool QQmlTypeLoader::Blob::isDebugging() const
     return typeLoader()->engine()->handle()->debugger() != nullptr;
 }
 
-bool QQmlTypeLoader::Blob::diskCacheEnabled() const
+bool QQmlTypeLoader::Blob::readCacheFile() const
 {
-    return typeLoader()->engine()->handle()->diskCacheEnabled();
+    return typeLoader()->engine()->handle()->diskCacheOptions()
+            & QV4::ExecutionEngine::DiskCache::QmlcRead;
+}
+
+bool QQmlTypeLoader::Blob::writeCacheFile() const
+{
+    return typeLoader()->engine()->handle()->diskCacheOptions()
+            & QV4::ExecutionEngine::DiskCache::QmlcWrite;
+}
+
+QQmlMetaType::CacheMode QQmlTypeLoader::Blob::aotCacheMode() const
+{
+    const QV4::ExecutionEngine::DiskCacheOptions options
+            = typeLoader()->engine()->handle()->diskCacheOptions();
+    if (!(options & QV4::ExecutionEngine::DiskCache::Aot))
+        return QQmlMetaType::RejectAll;
+    if (options & QV4::ExecutionEngine::DiskCache::AotByteCode)
+        return QQmlMetaType::AcceptUntyped;
+    return QQmlMetaType::RequireFullyTyped;
 }
 
 bool QQmlTypeLoader::Blob::qmldirDataAvailable(const QQmlRefPointer<QQmlQmldirData> &data, QList<QQmlError> *errors)
@@ -889,8 +971,9 @@ QQmlRefPointer<QQmlTypeData> QQmlTypeLoader::getType(const QUrl &unNormalizedUrl
         m_typeCache.insert(url, typeData);
         QQmlMetaType::CachedUnitLookupError error = QQmlMetaType::CachedUnitLookupError::NoError;
 
-        if (const QQmlPrivate::CachedQmlUnit *cachedUnit = typeData->diskCacheEnabled()
-                ? QQmlMetaType::findCachedCompilationUnit(typeData->url(), &error)
+        const QQmlMetaType::CacheMode cacheMode = typeData->aotCacheMode();
+        if (const QQmlPrivate::CachedQmlUnit *cachedUnit = (cacheMode != QQmlMetaType::RejectAll)
+                ? QQmlMetaType::findCachedCompilationUnit(typeData->url(), cacheMode, &error)
                 : nullptr) {
             QQmlTypeLoader::loadWithCachedUnit(typeData, cachedUnit, mode);
         } else {
@@ -901,16 +984,16 @@ QQmlRefPointer<QQmlTypeData> QQmlTypeLoader::getType(const QUrl &unNormalizedUrl
         // this was started Asynchronous, but we need to force Synchronous
         // completion now (if at all possible with this type of URL).
 
+#if QT_CONFIG(thread)
         if (!m_thread->isThisThread()) {
             // this only works when called directly from the UI thread, but not
             // when recursively called on the QML thread via resolveTypes()
 
             while (!typeData->isCompleteOrError()) {
-                unlock();
                 m_thread->waitForNextMessage();
-                lock();
             }
         }
+#endif
     }
 
     return typeData;
@@ -930,6 +1013,26 @@ QQmlRefPointer<QQmlTypeData> QQmlTypeLoader::getType(const QByteArray &data, con
     return QQmlRefPointer<QQmlTypeData>(typeData, QQmlRefPointer<QQmlTypeData>::Adopt);
 }
 
+void QQmlTypeLoader::injectScript(const QUrl &relativeUrl)
+{
+    LockHolder<QQmlTypeLoader> holder(this);
+
+    QQmlScriptBlob *blob = new QQmlScriptBlob(relativeUrl, this);
+    blob->initializeFromNative();
+    blob->m_isDone = true;
+    blob->m_data.setStatus(QQmlDataBlob::Complete);
+    m_scriptCache.insert(relativeUrl, blob);
+}
+
+QQmlRefPointer<QQmlScriptBlob> QQmlTypeLoader::injectedScript(const QUrl &relativeUrl)
+{
+    LockHolder<QQmlTypeLoader> holder(this);
+    const auto it = m_scriptCache.constFind(relativeUrl);
+    return (it != m_scriptCache.constEnd() && (*it)->hasScriptValue())
+            ? *it
+            : QQmlRefPointer<QQmlScriptBlob>();
+}
+
 /*!
 Return a QQmlScriptBlob for \a url.  The QQmlScriptData may be cached.
 */
@@ -950,8 +1053,9 @@ QQmlRefPointer<QQmlScriptBlob> QQmlTypeLoader::getScript(const QUrl &unNormalize
         m_scriptCache.insert(url, scriptBlob);
 
         QQmlMetaType::CachedUnitLookupError error = QQmlMetaType::CachedUnitLookupError::NoError;
-        if (const QQmlPrivate::CachedQmlUnit *cachedUnit = scriptBlob->diskCacheEnabled()
-                ? QQmlMetaType::findCachedCompilationUnit(scriptBlob->url(), &error)
+        const QQmlMetaType::CacheMode cacheMode = scriptBlob->aotCacheMode();
+        if (const QQmlPrivate::CachedQmlUnit *cachedUnit = (cacheMode != QQmlMetaType::RejectAll)
+                ? QQmlMetaType::findCachedCompilationUnit(scriptBlob->url(), cacheMode, &error)
                 : nullptr) {
             QQmlTypeLoader::loadWithCachedUnit(scriptBlob, cachedUnit);
         } else {
@@ -1231,6 +1335,11 @@ and qmldir information.
 */
 void QQmlTypeLoader::clearCache()
 {
+    // Pending messages typically hold references to the blobs they want to be delivered to.
+    // We don't want them anymore.
+    if (m_thread)
+        m_thread->discardMessages();
+
     for (TypeCache::Iterator iter = m_typeCache.begin(), end = m_typeCache.end(); iter != end; ++iter)
         (*iter)->release();
     for (ScriptCache::Iterator iter = m_scriptCache.begin(), end = m_scriptCache.end(); iter != end; ++iter)
@@ -1247,7 +1356,6 @@ void QQmlTypeLoader::clearCache()
     m_importDirCache.clear();
     m_importQmlDirCache.clear();
     m_checksumCache.clear();
-    QQmlMetaType::freeUnusedTypesAndCaches();
 }
 
 void QQmlTypeLoader::updateTypeCacheTrimThreshold()
@@ -1269,15 +1377,29 @@ void QQmlTypeLoader::trimCache()
             // typeData->m_compiledData may be set early on in the proccess of loading a file, so
             // it's important to check the general loading status of the typeData before making any
             // other decisions.
-            if (typeData->count() == 1 && (typeData->isError() || typeData->isComplete())
-                    && (!typeData->m_compiledData || typeData->m_compiledData->count() == 1)) {
-                // There are no live objects of this type
-                iter.value()->release();
-                iter = m_typeCache.erase(iter);
-                deletedOneType = true;
-            } else {
+            if (typeData->count() != 1 || (!typeData->isError() && !typeData->isComplete())) {
                 ++iter;
+                continue;
             }
+
+            const QQmlRefPointer<QV4::CompiledData::CompilationUnit> &compilationUnit
+                = typeData->m_compiledData;
+            if (compilationUnit) {
+                if (compilationUnit->count()
+                        > QQmlMetaType::countInternalCompositeTypeSelfReferences(
+                              compilationUnit) + 1) {
+                    ++iter;
+                    continue;
+                }
+
+                QQmlMetaType::unregisterInternalCompositeType(compilationUnit);
+                Q_ASSERT(compilationUnit->count() == 1);
+            }
+
+            // There are no live objects of this type
+            iter.value()->release();
+            iter = m_typeCache.erase(iter);
+            deletedOneType = true;
         }
 
         if (!deletedOneType)
