@@ -2,8 +2,12 @@
 // This code is governed by the BSD license found in the LICENSE file.
 
 /*---
-esid: prod-RegularExpressionBackslashSequence
+esid: sec-literals-regular-expression-literals
 info: |
+  RegularExpressionChar ::
+    RegularExpressionBackslashSequence
+
+
   RegularExpressionBackslashSequence ::
     \ RegularExpressionNonTerminator
 
@@ -18,6 +22,7 @@ info: |
 
 description: >
   A regular expression may not contain a <CR> as a SourceCharacter
+
 negative:
   phase: parse
   type: SyntaxError
@@ -25,5 +30,5 @@ negative:
 
 throw "Test262: This statement should not be evaluated.";
 
-/a\
+/a
 /
