@@ -1,11 +1,12 @@
 // Copyright (C) 2016 The Qt Company Ltd.
-// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 #ifndef TESTTYPES_H
 #define TESTTYPES_H
 
 #include <QtCore/qobject.h>
 #include <QtCore/qrect.h>
 #include <QtCore/qdatetime.h>
+#include <QtCore/qjsonarray.h>
 #include <QtGui/qtransform.h>
 #include <QtGui/qcolor.h>
 #include <QtGui/qvector2d.h>
@@ -18,6 +19,8 @@
 #include <QtQml/qqmlpropertyvaluesource.h>
 #include <QtQml/qqmlscriptstring.h>
 #include <QtQml/qqmlproperty.h>
+
+#include <private/qqmlcomponentattached_p.h>
 #include <private/qqmlcustomparser_p.h>
 
 QVariant myCustomVariantTypeConverter(const QString &data);
@@ -42,13 +45,86 @@ struct MyCustomVariantType
 };
 Q_DECLARE_METATYPE(MyCustomVariantType);
 
+
+class Group : public QObject
+{
+    Q_OBJECT
+
+    Q_PROPERTY(int value MEMBER value)
+
+public:
+    Group(QObject *parent = nullptr) : QObject(parent) {}
+    int value = 0;
+};
+
+
+struct GroupGadget
+{
+    Q_GADGET
+
+    Q_PROPERTY(int value MEMBER value)
+
+public:
+    friend bool operator==(GroupGadget g1, GroupGadget g2) { return g1.value == g2.value; }
+    friend bool operator!=(GroupGadget g1, GroupGadget g2) { return !(g1 == g2); }
+    int value = 0;
+};
+
+struct FakeDynamicObject : public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(QString value MEMBER value)
+
+public:
+    FakeDynamicObject() {}
+    QString value;
+};
+
+QT_BEGIN_NAMESPACE
+namespace QtPrivate {
+// don't do this at home – we override the meta-object which QMetaType collects for
+// FakeDynamicObject* properties
+template<>
+struct MetaObjectForType<FakeDynamicObject *, void>
+{
+    static const QMetaObject *metaObjectFunction(const QMetaTypeInterface *);
+};
+}
+QT_END_NAMESPACE
+
+class UnexposedBase : public QObject
+{
+    Q_OBJECT
+
+    Q_PROPERTY(Group *group MEMBER group)
+    Q_PROPERTY(GroupGadget groupGadget MEMBER groupGadget)
+    Q_PROPERTY(FakeDynamicObject *dynamic MEMBER dynamic)
+public:
+    UnexposedBase(QObject *parent = nullptr) : QObject(parent)
+    {
+        group = new Group(this);
+    }
+    Group *group;
+    GroupGadget groupGadget;
+    FakeDynamicObject *dynamic = nullptr;
+};
+
+class DerivedFromUnexposedBase : public UnexposedBase
+{
+    Q_OBJECT
+    QML_ELEMENT
+};
+
+
 class MyAttachedObject : public QObject
 {
     Q_OBJECT
     Q_PROPERTY(int value READ value WRITE setValue NOTIFY valueChanged)
     Q_PROPERTY(int value2 READ value2 WRITE setValue2)
+    Q_PROPERTY(int length MEMBER length)
+    Q_PROPERTY(QString name MEMBER name)
 public:
-    MyAttachedObject(QObject *parent) : QObject(parent), m_value(0), m_value2(0) {}
+    MyAttachedObject(QObject *parent) : QObject(parent), m_value(0), m_value2(0), name("attached") {}
 
     int value() const { return m_value; }
     void setValue(int v) { if (m_value != v) { m_value = v; emit valueChanged(); } }
@@ -62,6 +138,8 @@ signals:
 private:
     int m_value;
     int m_value2;
+    int length = 10;
+    QString name;
 };
 
 class SomethingUnknown : public QObject {
@@ -235,6 +313,7 @@ class MyTypeObject : public QObject
     Q_PROPERTY(qreal realProperty READ realProperty WRITE setRealProperty NOTIFY realPropertyChanged)
     Q_PROPERTY(double doubleProperty READ doubleProperty WRITE setDoubleProperty NOTIFY doublePropertyChanged)
     Q_PROPERTY(float floatProperty READ floatProperty WRITE setFloatProperty NOTIFY floatPropertyChanged)
+    Q_PROPERTY(qsizetype qsizetypeProperty READ qsizetypeProperty WRITE setQsizetypeProperty NOTIFY qsizetypePropertyChanged)
     Q_PROPERTY(QColor colorProperty READ colorProperty WRITE setColorProperty NOTIFY colorPropertyChanged)
     Q_PROPERTY(QDate dateProperty READ dateProperty WRITE setDateProperty NOTIFY datePropertyChanged)
     Q_PROPERTY(QTime timeProperty READ timeProperty WRITE setTimeProperty NOTIFY timePropertyChanged)
@@ -388,6 +467,15 @@ public:
         emit intPropertyChanged();
     }
 
+    qsizetype qsizetypePropertyValue;
+    qsizetype qsizetypeProperty() const {
+        return qsizetypePropertyValue;
+    }
+    void setQsizetypeProperty(const qsizetype &v) {
+        qsizetypePropertyValue = v;
+        emit qsizetypePropertyChanged();
+    }
+
     qreal realPropertyValue;
     qreal realProperty() const {
        return realPropertyValue;
@@ -604,6 +692,7 @@ signals:
     void byteArrayPropertyChanged();
     void uintPropertyChanged();
     void intPropertyChanged();
+    void qsizetypePropertyChanged();
     void realPropertyChanged();
     void doublePropertyChanged();
     void floatPropertyChanged();
@@ -1292,6 +1381,40 @@ public:
     }
 };
 
+class EnumPropsManyUnderlyingTypes : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+public:
+    enum  si8 : qint8 { ResolvedValue = 1};
+    enum  ui8 : quint8 {};
+    enum si16 : qint16 {};
+    enum ui16 : quint16 {};
+    enum ui64 : qint64 {};
+    enum si64 : quint64 {};
+    Q_ENUM(si8)
+    Q_ENUM(ui8)
+    Q_ENUM(si16)
+    Q_ENUM(ui16)
+    Q_ENUM(si64)
+    Q_ENUM(ui64)
+
+
+    Q_PROPERTY(si8  si8prop MEMBER si8prop)
+    Q_PROPERTY(ui8  ui8prop MEMBER ui8prop)
+    Q_PROPERTY(si16 si16prop MEMBER si16prop)
+    Q_PROPERTY(ui16 ui16prop MEMBER ui16prop)
+    Q_PROPERTY(si64 si64prop MEMBER si64prop)
+    Q_PROPERTY(ui64 ui64prop MEMBER ui64prop)
+
+    si8 si8prop = si8(0);
+    ui8 ui8prop = ui8(0);
+    si16 si16prop = si16(0);
+    ui16 ui16prop = ui16(0);
+    si64 si64prop = si64(0);
+    ui64 ui64prop = ui64(0);
+};
+
 Q_DECLARE_METATYPE(MyEnum2Class::EnumB)
 Q_DECLARE_METATYPE(MyEnum1Class::EnumA)
 Q_DECLARE_METATYPE(Qt::TextFormat)
@@ -1547,6 +1670,7 @@ class BareSingleton : public QObject
     Q_OBJECT
     QML_SINGLETON
     QML_ELEMENT
+    QML_ADDED_IN_VERSION(1, 0)
 
 public:
     BareSingleton(QObject *parent = nullptr) : QObject(parent)
@@ -1560,6 +1684,7 @@ class UncreatableSingleton : public QObject
     Q_OBJECT
     QML_SINGLETON
     QML_ELEMENT
+    QML_ADDED_IN_VERSION(1, 0)
 
 public:
     static UncreatableSingleton *instance();
@@ -1818,8 +1943,7 @@ public:
 
     int c()
     {
-        Q_UNREACHABLE();
-        return 1111;
+        Q_UNREACHABLE_RETURN(1111);
     }
 };
 
@@ -2213,22 +2337,725 @@ private:
 class BindableOnly : public QObject
 {
     Q_OBJECT
-    Q_PROPERTY(QByteArray data READ data WRITE setData BINDABLE dataBindable FINAL)
+    Q_PROPERTY(int score BINDABLE scoreBindable READ default WRITE default FINAL)
+    Q_PROPERTY(QByteArray data READ default WRITE default BINDABLE dataBindable FINAL)
     QML_ELEMENT
 public:
     BindableOnly(QObject *parent = nullptr)
         : QObject(parent)
+        , m_score(4)
     {}
-
+    QBindable<int> scoreBindable() { return QBindable<int>(&m_score); }
     QBindable<QByteArray> dataBindable() { return QBindable<QByteArray>(&m_data); }
 
-    QByteArray data() const { return m_data.value(); }
-    void setData(const QByteArray &newData) { m_data.setValue(newData); }
-
 private:
+    QProperty<int> m_score;
     QProperty<QByteArray> m_data;
 };
 
 void registerTypes();
 
+class AttachMe : public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(bool abc READ abc WRITE setAbc NOTIFY abcChanged)
+    QML_ANONYMOUS
+
+    bool m_abc;
+signals:
+    void abcChanged();
+
+public:
+    AttachMe(QObject *parent) : QObject(parent) { }
+    bool abc() const { return m_abc; }
+    void setAbc(bool abc) { m_abc = abc; }
+};
+
+class AnotherAttachMe : public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(QString anotherAbc READ anotherAbc WRITE setAnotherAbc NOTIFY anotherAbcChanged)
+    QML_ANONYMOUS
+
+    QString m_anotherAbc;
+signals:
+    void anotherAbcChanged();
+
+public:
+    AnotherAttachMe(QObject *parent) : QObject(parent) { }
+    QString anotherAbc() const { return m_anotherAbc; }
+    void setAnotherAbc(const QString &abc) { m_anotherAbc = abc; }
+};
+
+class OriginalQmlAttached : public QObject
+{
+    Q_OBJECT
+    QML_ATTACHED(AttachMe)
+    QML_ELEMENT
+
+public:
+    static AttachMe *qmlAttachedProperties(QObject *object) { return new AttachMe(object); }
+};
+
+class LeakingQmlAttached : public OriginalQmlAttached
+{
+    Q_OBJECT
+    QML_ELEMENT
+};
+
+class DerivedQmlAttached : public OriginalQmlAttached
+{
+    Q_OBJECT
+    QML_ELEMENT
+    QML_ATTACHED(AnotherAttachMe)
+
+public:
+    static AnotherAttachMe *qmlAttachedProperties(QObject *object)
+    {
+        return new AnotherAttachMe(object);
+    }
+};
+
+class OriginalSingleton : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+    QML_SINGLETON
+
+private:
+    Q_PROPERTY(QString abc READ abc WRITE setAbc NOTIFY abcChanged)
+
+    QString m_abc;
+signals:
+    void abcChanged(const QString &);
+
+public:
+    Q_INVOKABLE int mm() { return 5; }
+    QString abc() const { return m_abc; }
+    void setAbc(const QString &abc)
+    {
+        m_abc = abc;
+        emit abcChanged(abc);
+    }
+};
+
+class LeakingSingleton : public OriginalSingleton
+{
+    Q_OBJECT
+    QML_ELEMENT
+};
+
+class DerivedSingleton : public OriginalSingleton
+{
+    Q_OBJECT
+    QML_ELEMENT
+    QML_SINGLETON
+    Q_PROPERTY(QString anotherAbc READ anotherAbc WRITE setAnotherAbc NOTIFY anotherAbcChanged)
+
+    QString m_anotherAbc;
+signals:
+    void anotherAbcChanged(const QString &);
+
+public:
+    QString anotherAbc() const { return m_anotherAbc; }
+    void setAnotherAbc(const QString &abc)
+    {
+        m_anotherAbc = abc;
+        emit anotherAbcChanged(abc);
+    }
+};
+
+class Foreigner : public QObject
+{
+    Q_OBJECT
+
+private:
+    Q_PROPERTY(QString abc READ abc WRITE setAbc NOTIFY abcChanged)
+
+    QString m_abc;
+signals:
+    void abcChanged(const QString &);
+
+public:
+    QString abc() const { return m_abc; }
+    void setAbc(const QString &abc)
+    {
+        m_abc = abc;
+        emit abcChanged(abc);
+    }
+};
+
+class ForeignerForeign
+{
+    Q_GADGET
+    QML_ELEMENT
+    QML_FOREIGN(Foreigner)
+};
+
+class LeakingForeignerForeign : public QObject, public ForeignerForeign
+{
+    Q_OBJECT
+    QML_ELEMENT
+    Q_PROPERTY(QString anotherAbc READ anotherAbc WRITE setAnotherAbc NOTIFY anotherAbcChanged)
+
+    QString m_anotherAbc;
+signals:
+    void anotherAbcChanged(const QString &);
+
+public:
+    QString anotherAbc() const { return m_anotherAbc; }
+    void setAnotherAbc(const QString &abc)
+    {
+        m_anotherAbc = abc;
+        emit anotherAbcChanged(abc);
+    }
+};
+
+
+struct ForeignNamespace
+{
+    Q_GADGET
+public:
+    enum Abc { A, B, C, D };
+    Q_ENUM(Abc)
+};
+
+class ForeignNamespaceForeign
+{
+    Q_GADGET
+    QML_ELEMENT
+    QML_FOREIGN_NAMESPACE(ForeignNamespace)
+};
+
+class LeakingForeignNamespaceForeign : public QObject, public ForeignNamespaceForeign
+{
+    Q_OBJECT
+    QML_ELEMENT
+
+public:
+    enum AnotherAbc { D, C, B, A };
+    Q_ENUM(AnotherAbc)
+};
+
+struct ValueTypeWithLength
+{
+    Q_GADGET
+    QML_VALUE_TYPE(withLength)
+    QML_CONSTRUCTIBLE_VALUE
+
+    Q_PROPERTY(int length READ length CONSTANT)
+
+public:
+    ValueTypeWithLength() = default;
+    Q_INVOKABLE ValueTypeWithLength(int length) : m_length(length) {}
+    Q_INVOKABLE QString toString() const { return QStringLiteral("no"); }
+
+    int length() const { return m_length; }
+
+private:
+    int m_length = 19;
+};
+
+struct ValueTypeWithString
+{
+    Q_GADGET
+    QML_VALUE_TYPE(withString)
+    QML_CONSTRUCTIBLE_VALUE
+
+public:
+    Q_INVOKABLE ValueTypeWithString(const QString &v = QString()) : m_string(v) {}
+    QString toString() const { return m_string; }
+
+private:
+    QString m_string;
+};
+
+class GetterObject : public QObject {
+    Q_OBJECT
+    QML_ELEMENT
+    QML_SINGLETON
+public:
+    explicit GetterObject(QObject *parent = nullptr) : QObject{parent} {}
+
+    // always returns a 0 as uint64_t
+    Q_INVOKABLE uint64_t getFalse() const { return 0; }
+    Q_INVOKABLE uint64_t getTrue() const { return 1; }
+
+    Q_INVOKABLE quint64 getQFalse() const { return 0; }
+    Q_INVOKABLE quint64 getQTrue() const { return 1; }
+};
+
+class EnumProviderSingleton : public QObject {
+    Q_OBJECT
+
+public:
+    enum class Expected {
+        Value = 42
+    };
+    Q_ENUM(Expected)
+
+    EnumProviderSingleton(QObject* parent = nullptr) : QObject(parent) {}
+};
+
+class EnumProviderSingletonQml {
+    Q_GADGET
+    QML_FOREIGN(EnumProviderSingleton)
+    QML_NAMED_ELEMENT(EnumProviderSingleton)
+    QML_SINGLETON
+
+public:
+    static EnumProviderSingleton* create(QQmlEngine*, QJSEngine*) {
+        return new EnumProviderSingleton();
+    }
+
+private:
+    EnumProviderSingletonQml() = default;
+};
+
+
+
+namespace TypedEnums {
+Q_NAMESPACE
+QML_ELEMENT
+
+enum E8S : qint8 {
+    E8SA = std::numeric_limits<qint8>::min(),
+    E8SB = -5,
+    E8SC = -1,
+    E8SD = 0,
+    E8SE = 1,
+    E8SF = 5,
+    E8SG = std::numeric_limits<qint8>::max(),
+};
+Q_ENUM_NS(E8S);
+
+enum E8U : quint8 {
+    E8UA = 0,
+    E8UB = 1,
+    E8UC = 5,
+    E8UD = 1 << 7,
+    E8UE = std::numeric_limits<quint8>::max(),
+};
+Q_ENUM_NS(E8U);
+
+enum E16S : qint16 {
+    E16SA = std::numeric_limits<qint16>::min(),
+    E16SB = -5,
+    E16SC = -1,
+    E16SD = 0,
+    E16SE = 1,
+    E16SF = 5,
+    E16SG = std::numeric_limits<qint16>::max(),
+};
+Q_ENUM_NS(E16S);
+
+enum E16U : quint16 {
+    E16UA = 0,
+    E16UB = 1,
+    E16UC = 5,
+    E16UD = 1 << 15,
+    E16UE = std::numeric_limits<quint16>::max(),
+};
+Q_ENUM_NS(E16U);
+
+enum E32S : qint32 {
+    E32SA = std::numeric_limits<qint32>::min(),
+    E32SB = -5,
+    E32SC = -1,
+    E32SD = 0,
+    E32SE = 1,
+    E32SF = 5,
+    E32SG = std::numeric_limits<qint32>::max(),
+};
+Q_ENUM_NS(E32S);
+
+enum E32U : quint32 {
+    E32UA = 0,
+    E32UB = 1,
+    E32UC = 5,
+    E32UD = 1u << 31,
+    E32UE = std::numeric_limits<quint32>::max(),
+};
+Q_ENUM_NS(E32U);
+
+enum E64S : qint64 {
+    E64SA = std::numeric_limits<qint64>::min(),
+    E64SB = -5,
+    E64SC = -1,
+    E64SD = 0,
+    E64SE = 1,
+    E64SF = 5,
+    E64SG = std::numeric_limits<qint64>::max(),
+};
+Q_ENUM_NS(E64S);
+
+enum E64U : quint64 {
+    E64UA = 0,
+    E64UB = 1,
+    E64UC = 5,
+    E64UD = 1ull << 63,
+    E64UE = std::numeric_limits<quint64>::max(),
+};
+Q_ENUM_NS(E64U);
+}
+
+class GadgetWithEnums
+{
+    Q_GADGET
+    QML_VALUE_TYPE(gadgetWithEnums)
+    Q_PROPERTY(TypedEnums::E8S  e8s  MEMBER m_e8s);
+    Q_PROPERTY(TypedEnums::E8U  e8u  MEMBER m_e8u);
+    Q_PROPERTY(TypedEnums::E16S e16s MEMBER m_e16s);
+    Q_PROPERTY(TypedEnums::E16U e16u MEMBER m_e16u);
+    Q_PROPERTY(TypedEnums::E32S e32s MEMBER m_e32s);
+    Q_PROPERTY(TypedEnums::E32U e32u MEMBER m_e32u);
+    Q_PROPERTY(TypedEnums::E64S e64s MEMBER m_e64s);
+    Q_PROPERTY(TypedEnums::E64U e64u MEMBER m_e64u);
+public:
+    TypedEnums::E8S  m_e8s  = {};
+    TypedEnums::E8U  m_e8u  = {};
+    TypedEnums::E16S m_e16s = {};
+    TypedEnums::E16U m_e16u = {};
+    TypedEnums::E32S m_e32s = {};
+    TypedEnums::E32U m_e32u = {};
+    TypedEnums::E64S m_e64s = {};
+    TypedEnums::E64U m_e64u = {};
+private:
+    friend bool operator==(const GadgetWithEnums &a, const GadgetWithEnums &b)
+    {
+        return a.m_e8s == b.m_e8s && a.m_e8u == b.m_e8u && a.m_e16s == b.m_e16s
+                && a.m_e16u == b.m_e16u && a.m_e32s == b.m_e32s && a.m_e32u == b.m_e32u
+                && a.m_e64s == b.m_e64s && a.m_e64u == b.m_e64u;
+    }
+    friend bool operator!=(const GadgetWithEnums &a, const GadgetWithEnums &b)
+    {
+        return !(a == b);
+    }
+};
+
+class ObjectWithEnums : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+    Q_PROPERTY(TypedEnums::E8S  e8s  MEMBER m_e8s  NOTIFY changed);
+    Q_PROPERTY(TypedEnums::E8U  e8u  MEMBER m_e8u  NOTIFY changed);
+    Q_PROPERTY(TypedEnums::E16S e16s MEMBER m_e16s NOTIFY changed);
+    Q_PROPERTY(TypedEnums::E16U e16u MEMBER m_e16u NOTIFY changed);
+    Q_PROPERTY(TypedEnums::E32S e32s MEMBER m_e32s NOTIFY changed);
+    Q_PROPERTY(TypedEnums::E32U e32u MEMBER m_e32u NOTIFY changed);
+    Q_PROPERTY(TypedEnums::E64S e64s MEMBER m_e64s NOTIFY changed);
+    Q_PROPERTY(TypedEnums::E64U e64u MEMBER m_e64u NOTIFY changed);
+    Q_PROPERTY(GadgetWithEnums g MEMBER m_g NOTIFY changed);
+public:
+    ObjectWithEnums(QObject *parent = nullptr) : QObject(parent) {}
+    TypedEnums::E8S  m_e8s  = {};
+    TypedEnums::E8U  m_e8u  = {};
+    TypedEnums::E16S m_e16s = {};
+    TypedEnums::E16U m_e16u = {};
+    TypedEnums::E32S m_e32s = {};
+    TypedEnums::E32U m_e32u = {};
+    TypedEnums::E64S m_e64s = {};
+    TypedEnums::E64U m_e64u = {};
+    GadgetWithEnums m_g;
+Q_SIGNALS:
+    void changed();
+};
+
+struct UnregisteredValueBaseType
+{
+    int foo = 12;
+};
+
+struct UnregisteredValueDerivedType: public UnregisteredValueBaseType
+{
+    int bar = 13;
+};
+
+struct GadgetedValueBaseType
+{
+    Q_GADGET
+    int foo = 12;
+};
+
+struct GadgetedValueDerivedType: public GadgetedValueBaseType
+{
+    Q_GADGET
+    int bar = 13;
+};
+
+class UnregisteredValueTypeHandler: public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+public:
+    int consumed = 0;
+    int gadgeted = 0;
+
+public slots:
+    UnregisteredValueBaseType produce() { return UnregisteredValueBaseType(); }
+    UnregisteredValueDerivedType produceDerived() { return UnregisteredValueDerivedType(); }
+    void consume(UnregisteredValueBaseType) { ++consumed; }
+
+    GadgetedValueDerivedType produceGadgeted() { return GadgetedValueDerivedType(); }
+    void consume(GadgetedValueBaseType) { ++gadgeted; }
+};
+
+class Greeter : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+
+public:
+    Greeter(QObject *parent = nullptr) : QObject(parent) {}
+
+    Q_INVOKABLE void greet()
+    {
+        qDebug().noquote() << objectName() << "says hello";
+    }
+
+    Q_INVOKABLE void sum(int a, int b)
+    {
+        qDebug().noquote() << objectName() << QString("says %1 + %2 = %3").arg(a).arg(b).arg(a + b);
+    }
+};
+
+class Attachment : public QObject {
+    Q_OBJECT
+public:
+    Attachment(QObject *parent = nullptr) : QObject(parent) {}
+};
+
+class AttachedInCtor : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+    QML_ATTACHED(Attachment)
+
+public:
+    AttachedInCtor(QObject *parent = nullptr)
+        : QObject(parent)
+    {
+        attached = qmlAttachedPropertiesObject<AttachedInCtor>(this, true);
+    }
+
+    static Attachment *qmlAttachedProperties(QObject *object) {
+        return new Attachment(object);
+    }
+
+    QObject *attached = nullptr;
+};
+
+class BirthdayParty : public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(QQmlListProperty<QObject> guests READ guests)
+    Q_CLASSINFO("DefaultProperty", "guests")
+    QML_ELEMENT
+
+public:
+    using QObject::QObject;
+    QQmlListProperty<QObject> guests() { return {this, &m_guests}; }
+    qsizetype guestCount() const { return m_guests.count(); }
+    QObject *guest(qsizetype i) const { return m_guests.at(i); }
+
+private:
+    QList<QObject *> m_guests;
+};
+
+class ByteArrayReceiver : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+
+public:
+    QList<QByteArray> byteArrays;
+
+    Q_INVOKABLE void byteArrayTest(const QByteArray &ba)
+    {
+        byteArrays.push_back(ba);
+    }
+};
+
+class CounterAttachedBaseType: public QObject
+{
+    Q_OBJECT
+    QML_ANONYMOUS
+    Q_PROPERTY (int value READ value NOTIFY valueChanged)
+
+public:
+    CounterAttachedBaseType(QObject *parent = nullptr) : QObject(parent) {}
+
+    int value() { return m_value; }
+    Q_SIGNAL void valueChanged();
+
+protected:
+    int m_value = 98;
+};
+
+
+class CounterAttachedType: public CounterAttachedBaseType
+{
+    Q_OBJECT
+    QML_ANONYMOUS
+
+public:
+    CounterAttachedType(QObject *parent = nullptr) : CounterAttachedBaseType(parent) {}
+
+    Q_INVOKABLE void increase() {
+        ++m_value;
+        Q_EMIT valueChanged();
+    }
+};
+
+class Counter : public QObject
+{
+    Q_OBJECT
+    QML_ATTACHED(CounterAttachedBaseType)
+    QML_ELEMENT
+
+public:
+    static CounterAttachedBaseType *qmlAttachedProperties(QObject *o)
+    {
+        return new CounterAttachedType(o);
+    }
+};
+
+
+class Singleton: public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(EnumType enumProperty READ enumProperty CONSTANT)
+    Q_CLASSINFO("RegisterEnumClassesUnscoped", "false")
+    QML_ELEMENT
+    QML_SINGLETON
+public:
+    explicit Singleton(QObject* parent = nullptr) : QObject(parent) {}
+    enum class EnumType {
+        EnumValue1,
+        EnumValue2
+    };
+    Q_ENUM(EnumType);
+    EnumType enumProperty() const {
+        return EnumType::EnumValue2;
+    }
+};
+
+class NonSingleton: public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(EnumType enumProperty READ enumProperty CONSTANT)
+    Q_CLASSINFO("RegisterEnumClassesUnscoped", "false")
+    QML_ELEMENT
+public:
+    explicit NonSingleton(QObject* parent = nullptr) : QObject(parent) {}
+    enum class EnumType {
+        EnumValue1,
+        EnumValue2
+    };
+    Q_ENUM(EnumType);
+    EnumType enumProperty() const {
+        return EnumType::EnumValue2;
+    }
+};
+
+class TypeWithQJsonArrayProperty : public QObject {
+    Q_OBJECT
+    QML_ELEMENT
+
+    Q_PROPERTY(QJsonArray jsonArray READ jsonArray WRITE setJsonArray NOTIFY jsonArrayChanged)
+
+public:
+    TypeWithQJsonArrayProperty(QObject *parent = nullptr) : QObject(parent) {}
+
+    const QJsonArray& jsonArray() { return m_jsonArray; }
+    void setJsonArray(const QJsonArray& a) { m_jsonArray = a; }
+
+signals:
+    void jsonArrayChanged();
+
+private:
+    QJsonArray m_jsonArray;
+};
+
+class InvokableSingleton : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+    QML_SINGLETON
+public:
+    InvokableSingleton() = default;
+    Q_INVOKABLE InvokableSingleton(int a, QObject *parent) : QObject(parent), m_a(a) {}
+
+    int m_a = 0;
+};
+
+class InvokableExtension : public QObject
+{
+    Q_OBJECT
+public:
+    Q_INVOKABLE InvokableExtension(QObject *parent = nullptr) : QObject(parent) {}
+};
+
+class InvokableExtended : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+    QML_EXTENDED(InvokableExtension)
+
+public:
+    Q_INVOKABLE InvokableExtended() = default;
+};
+
+class InvokableUncreatable : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+    QML_UNCREATABLE("no")
+
+public:
+    Q_INVOKABLE InvokableUncreatable() = default;
+};
+
+class InvokableValueType
+{
+    Q_GADGET
+    QML_VALUE_TYPE(vv)
+public:
+    Q_INVOKABLE InvokableValueType() = default;
+    Q_INVOKABLE InvokableValueType(const QString &s) : m_s(s) {}
+    QString m_s;
+};
+
+class NestedVectors : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+public:
+    NestedVectors(QObject *parent = nullptr) : QObject(parent)
+    {
+        std::vector<int> data;
+        data.push_back(1);
+        data.push_back(2);
+        data.push_back(3);
+        m_list.push_back(data);
+        data.clear();
+        data.push_back(4);
+        data.push_back(5);
+        m_list.push_back(data);
+    }
+
+    Q_INVOKABLE std::vector<std::vector<int>> getList()
+    {
+        return m_list;
+    }
+
+    Q_INVOKABLE void setList(std::vector<std::vector<int>> list)
+    {
+        m_list = list;
+    }
+
+private:
+    std::vector<std::vector<int>> m_list;
+};
+
 #endif // TESTTYPES_H
